# ----------------------------------------------------------------------------
# Copyright (c) 2017-2022, Ben Kaehler.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from qiime2.plugin import (Plugin, List, Str, Float, Bool, Int, Citations,
                           Range, Choices)
from q2_types.feature_table import FeatureTable, RelativeFrequency, Frequency
from q2_types.feature_data import FeatureData, Taxonomy, Sequence
from q2_feature_classifier._taxonomic_classifier import TaxonomicClassifier

import q2_clawback

citations = Citations.load('citations.bib', package='q2_clawback')
plugin = Plugin(
    name='clawback',
    version=q2_clawback.__version__,
    website='https://github.com/BenKaehler/q2-clawback',
    package='q2_clawback',
    citations=[citations['bokulich2018optimizing'],
               citations['kaehler2019species']],
    description=('This QIIME 2 plugin provides support for generating '
                 'generating class weights for use with the '
                 'feature-classifier'),
    short_description='CLAss Weight Assembler plugin.'
)

plugin.visualizers.register_function(
    function=q2_clawback.summarize_Qiita_metadata_category_and_contexts,
    inputs={},
    parameters={'category': Str},
    name='Fetch Qiita sample types and contexts',
    description='Display of counts of samples grouped by category and context',
    parameter_descriptions={
        'category': 'Metadata key over which to summarize sample counts'
    }
)

plugin.methods.register_function(
    function=q2_clawback.sequence_variants_from_samples,
    inputs={'samples': FeatureTable[Frequency]},
    parameters=None,
    outputs=[('sequences', FeatureData[Sequence])],
    name='Extract sequence variants from a feature table',
    description=('Extract sequence variants from a feature table, '
                 'if the feature table observations are labelled by sequence '
                 'variant'),
    input_descriptions={
        'samples': 'Feature table from which to extract sequence variants. '
                   'Features must be sequence variants.'
    },
    output_descriptions={
        'sequences': 'Features from the table as sequences'
    }
)

_generate_class_weights_input_descriptions = {
    'reference_taxonomy': 'Taxonomy that will be used when the weights from '
                          'this method are used to train a taxonomic '
                          'classifier',
    'reference_sequences': 'Sequences that will be used when the weights from '
                           'this method are used to train a taxonomic '
                           'classifier'
}

_generate_class_weights_parameter_descriptions = {
    'unobserved_weight': 'Resulting weights are a weighted sum of the weights '
                         'aggregated from samples and uniform weights across '
                         'all reference taxa (because zero weights are not '
                         'ok). unobserved_weight is the weight given to the '
                         'uniform weights in the sum',
    'normalise': 'Whether to normalise counts within samples prior to weight '
                 'aggregation',
    'allow_weight_outside_reference': 'Whether to allow sample weight for '
                                      'taxa that do not exist in the '
                                      'reference taxonomy. Such taxa are '
                                      'ignored if True'
}

_generate_class_weights_output_descriptions = {
    'class_weight': 'Taxonomic weights for use training taxonomic classifiers'
}


plugin.methods.register_function(
    function=q2_clawback.generate_class_weights,
    inputs={'reference_taxonomy': FeatureData[Taxonomy],
            'reference_sequences': FeatureData[Sequence],
            'samples': FeatureTable[Frequency],
            'taxonomy_classification': FeatureData[Taxonomy]},
    parameters={'unobserved_weight': Float,
                'normalise': Bool,
                'allow_weight_outside_reference': Bool},
    outputs=[('class_weight', FeatureTable[RelativeFrequency])],
    name='Generate class weights from a set of samples',
    description=('Generate class weights for use with a taxonomic classifier '
                 'from a set of existing observations'),
    input_descriptions={
        'samples': 'Samples from which to assemble weights',
        'taxonomy_classification': 'Taxonomy classification that maps the '
                                   'features in samples to taxa',
        **_generate_class_weights_input_descriptions
    },
    parameter_descriptions=_generate_class_weights_parameter_descriptions,
    output_descriptions=_generate_class_weights_output_descriptions
)

_fetch_Qiita_samples_parameter_descriptions = {
    'metadata_key': 'Fetch samples where this metadata key matches any '
                    'of the values in the metada value list',
    'metadata_value': 'Fetch samples where the metadata key matches this '
                      'value',
    'context': 'The redbiom context. Should be a context that contains '
               'SVs. Something like Deblur-Illumina-16S-V4-150nt-XXXXXX'
}


plugin.methods.register_function(
    function=q2_clawback.fetch_Qiita_samples,
    inputs={},
    parameters={'metadata_value': List[Str],
                'context': Str,
                'metadata_key': Str},
    outputs=[('samples', FeatureTable[Frequency])],
    name='Fetch feature counts for a collection of samples',
    description=('Fetch feature counts for a collection of samples, '
                 'preferably with SVs for OTU ids'),
    parameter_descriptions=_fetch_Qiita_samples_parameter_descriptions,
    output_descriptions={
        'samples': 'All the samples matching the query found in Qiita'
    }
)

plugin.pipelines.register_function(
    function=q2_clawback.assemble_weights_from_Qiita,
    inputs={'classifier': TaxonomicClassifier,
            'reference_taxonomy': FeatureData[Taxonomy],
            'reference_sequences': FeatureData[Sequence]},
    parameters={
        'metadata_value': List[Str],
        'context': Str,
        'unobserved_weight': Float,
        'normalise': Bool,
        'metadata_key': Str,
        'n_jobs': Int,
        'reads_per_batch': Int % Range(1, None) | Str % Choices(['auto']),
        'allow_weight_outside_reference': Bool},
    outputs=[('class_weight', FeatureTable[RelativeFrequency])],
    name='Assemble weights from Qiita for use with q2-feature-classifier',
    description=('Download SV results from Qiita, classify the SVs, use the '
                 'result to collate class weights'),
    input_descriptions={
        'classifier': 'Taxonomic classifier to be used to classify SVs prior '
                      'to taxonomic weight assembly',
        **_generate_class_weights_input_descriptions
    },
    parameter_descriptions={
        'n_jobs': 'The maximum number of concurrently worker processes. If -1 '
                  'all CPUs are used. If 1 is given, no parallel computing '
                  'code is used at all, which is useful for debugging. For '
                  'n_jobs below -1, (n_cpus + 1 + n_jobs) are used. Thus for '
                  'n_jobs = -2, all CPUs but one are used.',
        'reads_per_batch': 'Number of reads to process in each batch. If 0, '
                           'this parameter is autoscaled to '
                           'min( number of query sequences / n_jobs, 20000).',
        **_fetch_Qiita_samples_parameter_descriptions,
        **_generate_class_weights_parameter_descriptions
    },
    output_descriptions=_generate_class_weights_output_descriptions
)



# ----------------------------------------------------------------------------
# Copyright (c) 2017-2022, Ben Kaehler.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------
import os

import pkg_resources
import biom
import q2templates
import redbiom.fetch
import redbiom.summarize
import redbiom.search
from pandas import Series, DataFrame
from numpy import array
from skbio import DNA
from q2_types.feature_data import DNAIterator

TEMPLATES = pkg_resources.resource_filename('q2_clawback', 'assets')


def sequence_variants_from_samples(samples: biom.Table) -> DNAIterator:
    seqs = (DNA(s, metadata={'id': s})
            for s in samples.ids(axis='observation'))
    return DNAIterator(seqs)


def _fetch_Qiita_summaries(category='sample_type'):
    md = redbiom.fetch.category_sample_values(category)
    counts = md.value_counts(ascending=False)
    caches = redbiom.summarize.contexts()[['ContextName', 'SamplesWithData']]
    caches = caches.sort_values(by='SamplesWithData', ascending=False)
    return counts, caches


def summarize_Qiita_metadata_category_and_contexts(
        output_dir: str = None, category: str = 'sample_type'):
    counts, caches = _fetch_Qiita_summaries(category=category)
    counts = counts.to_frame()
    counts = DataFrame({category: counts.index, 'count': counts.values.T[0]},
                       columns=[category, 'count'])
    sample_types = q2templates.df_to_html(counts, bold_rows=False, index=False)
    contexts = q2templates.df_to_html(caches, index=False)
    title = 'Available in Qiita'
    index = os.path.join(TEMPLATES, 'index.html')
    q2templates.render(index, output_dir, context={
        'title': title,
        'sample_types': sample_types,
        'contexts': contexts})


def fetch_Qiita_samples(metadata_value: list, context: str,
                        metadata_key: str = 'sample_type') -> biom.Table:
    query = "where " + metadata_key + " == '"
    query += ("' or " + metadata_key + " == '").join(metadata_value)
    query += "'"
    sample_ids = redbiom.search.metadata_full(query, False)
    samples, ambig = redbiom.fetch.data_from_samples(context, sample_ids)
    return samples


def generate_class_weights(
        reference_taxonomy: Series, reference_sequences: DNAIterator,
        samples: biom.Table, taxonomy_classification: DataFrame,
        unobserved_weight: float = 1e-6, normalise: bool = False,
        allow_weight_outside_reference: bool = False) \
        -> biom.Table:
    weights = {reference_taxonomy[seq.metadata['id']]: 0.
               for seq in reference_sequences}
    if normalise:
        samples.norm()

    tax_map = taxonomy_classification['Taxon']
    try:
        taxa = [tax_map[s] for s in samples.ids(axis='observation')]
    except KeyError as s:
        raise ValueError(str(s) + ' not in taxonomy_classification')
    if not allow_weight_outside_reference and not set(taxa).issubset(weights):
        raise ValueError(
            'taxonomy_classification does not match reference_taxonomy')

    for taxon, count in zip(taxa, samples.sum('observation')):
        if taxon in weights:
            weights[taxon] += count
    taxa, weights = zip(*weights.items())
    weights = array(weights)
    weights /= weights.sum()
    weights = \
        (1. - unobserved_weight) * weights + unobserved_weight / len(weights)
    weights /= weights.sum()

    return biom.Table(weights[None].T, taxa, sample_ids=['Weight'])


def assemble_weights_from_Qiita(
        ctx, classifier, reference_taxonomy, reference_sequences,
        metadata_value, context, unobserved_weight=1e-6, normalise=False,
        metadata_key='sample_type', n_jobs=1, reads_per_batch='auto',
        allow_weight_outside_reference=False):
    samples, = ctx.get_action('clawback', 'fetch_Qiita_samples')(
        metadata_value=metadata_value, context=context,
        metadata_key=metadata_key)

    reads, = ctx.get_action('clawback', 'sequence_variants_from_samples')(
        samples=samples)

    classification, = ctx.get_action('feature-classifier', 'classify_sklearn')(
        reads=reads, classifier=classifier, confidence='disable',
        n_jobs=n_jobs, reads_per_batch=reads_per_batch)

    return tuple(ctx.get_action('clawback', 'generate_class_weights')(
        reference_taxonomy=reference_taxonomy,
        reference_sequences=reference_sequences,
        samples=samples, taxonomy_classification=classification,
        unobserved_weight=unobserved_weight, normalise=normalise,
        allow_weight_outside_reference=allow_weight_outside_reference))



# ----------------------------------------------------------------------------
# Copyright (c) 2017-2022, Ben Kaehler.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

from ._version import get_versions
from ._clawback import (summarize_Qiita_metadata_category_and_contexts,
                        fetch_Qiita_samples,
                        sequence_variants_from_samples,
                        generate_class_weights,
                        assemble_weights_from_Qiita)

__all__ = ['summarize_Qiita_metadata_category_and_contexts',
           'sequence_variants_from_samples',
           'fetch_Qiita_samples',
           'generate_class_weights',
           'assemble_weights_from_Qiita']

__version__ = get_versions()['version']
del get_versions



# ----------------------------------------------------------------------------
# Copyright (c) 2017-2022, Ben Kaehler.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE, distributed with this software.
# ----------------------------------------------------------------------------

import tempfile
from warnings import filterwarnings

from qiime2 import Artifact
from qiime2.plugin.testing import TestPluginBase
from qiime2.plugins.feature_classifier.methods import \
    fit_classifier_naive_bayes
from qiime2.plugins.clawback.pipelines import assemble_weights_from_Qiita
from qiime2.plugins.clawback.methods import (
    sequence_variants_from_samples, generate_class_weights)
from qiime2.plugins.feature_classifier.methods import classify_sklearn
from biom import Table
from q2_types.feature_data import DNAIterator
from pandas import DataFrame, Series

import q2_clawback


class ClawbackTestPluginBase(TestPluginBase):
    package = 'q2_clawback.tests'

    def setUp(self):
        self.temp_dir = tempfile.TemporaryDirectory(
            prefix='q2-clawback-test-temp-')

        filterwarnings('ignore', 'The TaxonomicClassifier ', UserWarning)
        self.reads = Artifact.import_data(
            'FeatureData[Sequence]',
            self.get_data_path('se-dna-sequences.fasta'))
        self.taxonomy = Artifact.import_data(
            'FeatureData[Taxonomy]', self.get_data_path('taxonomy.tsv'))
        classifier = fit_classifier_naive_bayes(self.reads, self.taxonomy)
        self.classifier = classifier.classifier

    def tearDown(self):
        self.temp_dir.cleanup()


class ClawbackTests(ClawbackTestPluginBase):
    def _check_weights(self, weights):
        seqs = self.reads.view(DNAIterator)
        tax = self.taxonomy.view(DataFrame)['Taxon']
        taxa = set(tax[s.metadata['id']]
                   for s in seqs if s.metadata['id'] in tax)
        self.assertEqual(taxa, set(weights.ids(axis='observation')))
        self.assertEqual(weights.shape, (len(taxa), 1))
        weights = weights.data('Weight')
        self.assertTrue((weights > 0.).all())
        self.assertTrue((weights < 1.).all())
        self.assertAlmostEqual(weights.sum(), 1.)

    def test_assemble_weights_from_Qiita(self):
        counts, caches = q2_clawback._clawback._fetch_Qiita_summaries()

        sample_type = 'Tears'
        self.assertTrue(hasattr(counts, sample_type))
        for context in caches.ContextName:
            if context.startswith('Deblur') and 'Illumina-16S-V4' in context:
                break
        else:
            self.assertTrue(False)

        weights = assemble_weights_from_Qiita(
            self.classifier, self.taxonomy, self.reads, [sample_type], context)
        weights = weights[0].view(Table)

        self._check_weights(weights)

    def test_allow_weight_outside_reference(self):
        samples = Artifact.import_data(
            'FeatureTable[Frequency]', self.get_data_path('tears.biom'))
        sample_reads = sequence_variants_from_samples(samples).sequences
        taxonomy = classify_sklearn(
            reads=sample_reads, classifier=self.classifier,
            confidence='disable', n_jobs=2, reads_per_batch=10)
        taxonomy = taxonomy.classification.view(Series)
        taxonomy['TACGGAGGGTGCAAGCGTTAATCGGAATTACTGGGCGTAAAGCGCGCGTAGGCGGCTAGG'
                 'TCAGTCAGATGTGAAAGCCCCGGGCTTAACCTGGGAATTG'] = 'Not a taxon'
        taxonomy = Artifact.import_data('FeatureData[Taxonomy]', taxonomy)
        with self.assertRaises(ValueError):
            weights = generate_class_weights(
                self.taxonomy, self.reads, samples, taxonomy)
        weights = generate_class_weights(
            self.taxonomy, self.reads, samples, taxonomy,
            allow_weight_outside_reference=True)
        weights = weights.class_weight.view(Table)

        self._check_weights(weights)
